@@ -95,7 +95,7 @@ TODO:
 import sys
 import time
 import collections
-
+from itertools import islice
 
 __all__ = [
     'ProgIter',
@@ -103,6 +103,9 @@ __all__ = [
 
 default_timer = time.perf_counter
 
+# A measurment takes place at a given iteration and posixtime.
+Measurement = collections.namedtuple('Measurement', ['idx', 'time'])
+
 
 CLEAR_BEFORE = '\r'
 AT_END = '\n'
@@ -243,6 +246,18 @@ class _BackwardsCompat(object):
         """ alias of desc """
         return self.desc
 
+    def start(self):  # nocover
+        """
+        Alias of :func:`ProgIter.begin`
+        """
+        return self.begin()
+
+    def stop(self):  # nocover
+        """
+        Alias of :func:`ProgIter.end`
+        """
+        return self.end()
+
 
 class ProgIter(_TQDMCompat, _BackwardsCompat):
     """
@@ -250,7 +265,7 @@ class ProgIter(_TQDMCompat, _BackwardsCompat):
 
     ProgIter is an alternative to `tqdm`. ProgIter implements much of the
     tqdm-API.  The main difference between `ProgIter` and `tqdm` is that
-    ProgIter does not use threading where as `tqdm` does.
+    ProgIter does not use threading whereas `tqdm` does.
 
     Attributes:
         iterable (List | Iterable):
@@ -267,10 +282,6 @@ class ProgIter(_TQDMCompat, _BackwardsCompat):
             How many iterations to wait between messages.
             Defaults to 1.
 
-        adjust (bool):
-            if True freq is adjusted based on time_thresh
-            Defaults to True.
-
         eta_window (int):
             number of previous measurements to use in eta calculation, default=64
 
@@ -318,6 +329,13 @@ class ProgIter(_TQDMCompat, _BackwardsCompat):
             2: enabled=True with clearline=False and adjust=True,
             3: enabled=True with clearline=False and adjust=False
 
+        homogeneous (bool | str):
+            Indicate if the iterable is likely to take a uniform or homogeneous
+            amount of time per iteration. When True we can enable a speed
+            optimization. When False, the time estimates are more accurate.
+            Default to "auto", which attempts to determine if it is safe to use
+            True. Has no effect if ``adjust`` is False.
+
     Note:
         Either use ProgIter in a with statement or call prog.end() at the end
         of the computation if there is a possibility that the entire iterable
@@ -325,7 +343,7 @@ class ProgIter(_TQDMCompat, _BackwardsCompat):
 
     Note:
         ProgIter is an alternative to `tqdm`.  The main difference between
-        `ProgIter` and `tqdm` is that ProgIter does not use threading where as
+        `ProgIter` and `tqdm` is that ProgIter does not use threading whereas
         `tqdm` does.  `ProgIter` is simpler than `tqdm` and thus more stable in
         certain circumstances.
 
@@ -348,7 +366,8 @@ class ProgIter(_TQDMCompat, _BackwardsCompat):
                  initial=0, eta_window=64, clearline=True, adjust=True,
                  time_thresh=2.0, show_times=True, show_wall=False,
                  enabled=True, verbose=None, stream=None, chunksize=None,
-                 rel_adjust_limit=4.0, **kwargs):
+                 rel_adjust_limit=4.0, homogeneous='auto', timer=None,
+                 **kwargs):
         """
         Note:
             See attributes for arg information
@@ -413,6 +432,13 @@ class ProgIter(_TQDMCompat, _BackwardsCompat):
         self.started = False
         self.finished = False
 
+        if timer is None:
+            timer = default_timer
+        self._timer = timer
+
+        self.homogeneous = homogeneous
+        self._likely_homogeneous = None
+
         # indicates if the cursor is currently at the start of a line (True) or
         # if characters have been written with no newline yet.
         self._cursor_at_newline = True
@@ -465,22 +491,177 @@ class ProgIter(_TQDMCompat, _BackwardsCompat):
         """
         self.extra = extra
 
+    def _reset_internals(self):
+        """
+        Initialize all variables used in the internal state
+        """
+        # Prepare for iteration
+        if self.total is None:
+            self.total = _infer_length(self.iterable)
+
+        # Track the total time up to the most recent measurement.
+        self._total_seconds = 0
+
+        # Track the current iteration we are on
+        self._iter_idx = self.initial
+
+        # Track the last time we displayed a message
+        self._display_measurement = Measurement(-1, -1)
+
+        # Track the most recent iteration/time a measurement was made
+        self._curr_measurement = Measurement(self.initial, 0)
+
+        # Track the number of iterations and time between the last two measurements
+        self._measure_countdelta = -1
+        self._measure_timedelta = self.time_thresh
+        self._display_timedelta = self.time_thresh
+        self._next_measure_idx = self._iter_idx + self.freq
+
+        # Primary estimates
+        self._est_seconds_left = None
+        self._iters_per_second = 0.0
+
+        # hack flag that should be refactored and removed used to ensure the
+        # first message after begin is displayed.
+        self._force_next_display = False
+
+        self._update_message_template()
+
+    def begin(self):
+        """
+        Initializes information used to measure progress
+
+        This only needs to be used if this ProgIter is not wrapping an iterable.
+        Does nothing if this ProgIter is disabled.
+
+        Returns:
+            ProgIter:
+                a chainable self-reference
+        """
+        if not self.enabled:
+            return
+
+        self._reset_internals()
+
+        # Time progress was initialized
+        self._start_time = self._timer()
+        # Last time measures were updated
+        self._curr_measurement = Measurement(self._iter_idx, self._start_time)
+
+        # use last few times to compute a more stable average rate
+        if self.eta_window is not None:
+            self._measurements = collections.deque([
+                self._curr_measurement
+            ], maxlen=self.eta_window)
+        else:
+            self._measurements = collections.deque([
+                self._curr_measurement
+            ], maxlen=2)
+
+        # self._cursor_at_newline = True
+        self._cursor_at_newline = not self.clearline
+        self.started = True
+        self.finished = False
+
+        self._tryflush()
+        self.display_message()
+
+        # The start message isn't very helpful.
+        # If we enable this we could force the first iteration.
+        self._force_next_display = self.freq == 1
+        return self
+
+    def end(self):
+        """
+        Signals that iteration has ended and displays the final message.
+
+        This only needs to be used if this ProgIter is not wrapping an
+        iterable.  Does nothing if this ProgIter object is disabled or has
+        already finished.
+        """
+        if not self.enabled or self.finished:
+            return
+        # Write the final progress line if it was not written in the loop
+        if self._iter_idx != self._display_measurement.idx:
+            self._measure_time()
+            self._est_seconds_left = 0
+            self.display_message()
+        self.ensure_newline()
+        self._cursor_at_newline = True
+        self.finished = True
+
     def _iterate(self):
         """ iterates with progress """
         if not self.started:
             self.begin()
         # Wrap input sequence in a generator
-        for self._iter_idx, item in enumerate(self.iterable, start=self.initial + 1):
-            yield item
-            # Call the body of step to reduce overyead
-            # self.step(0)  # inc is 0 because we already updated
-            if (self._iter_idx) % self.freq == 0:
-                # update progress information every so often
-                self._update_measurements()
-                self._update_estimates()
-                self.display_message()
+        gen = enumerate(self.iterable, start=self.initial + 1)
+        # Iterating is performance sensitive, so separate both cases - where
+        # 'freq' is used and checks can be fast, and where 'adjust' is used and
+        # checks need more calculation. This is worth duplicating code for.
+
+        if self.adjust:
+            homogeneous = self.homogeneous
+            if homogeneous == 'auto':
+                yield from self._homogeneous_check(gen)
+                homogeneous = self._likely_homogeneous
+
+            if homogeneous:
+                use_fast_path = True
+            else:
+                use_fast_path = False
+                # Slow path where we do checks every iteration.
+                for self._iter_idx, item in gen:
+                    yield item
+                    self._slow_path_step_body()
+        else:
+            use_fast_path = True
+
+        if use_fast_path:
+            # In the fast path we only check the time every `freq` iterations.
+            for self._iter_idx, item in gen:
+                yield item
+                if self._force_next_display or self._iter_idx >= self._next_measure_idx:
+                    self._measure_time()
+                    if self._force_next_display or (self._display_timedelta >= self.time_thresh):
+                        self.display_message()
+
         self.end()
 
+    def _homogeneous_check(self, gen):
+        # NOTE: We could have a more complex heuristic with negligable
+        # overhead and more robustness that checks every n iterations
+        # that such that the time call overhead would be negligable.
+        # To do this we would need a semi-fast mode that does the fast
+        # mode for a fixed number of iterations and then rechecks the
+        # slow mode. Or something like that.
+        # NOTE: We could also try to find a pseudo property to check to
+        # see if things are changing. Is this faster than a call to
+        # time.time?
+        # Take a few steps in the slow path and then check to see
+        # if we should continue or do go down the fast path.
+        num_initial_steps = 5
+
+        # A call to time is 50ns, we can accept the overhead if it
+        # is only .01% of the total loop time
+        overhead_threshold = 50e-9 * 10_000
+
+        slowest = 0
+        for self._iter_idx, item in islice(gen, num_initial_steps):
+            yield item
+            self._slow_path_step_body()
+            slowest = max(slowest, self._measure_timedelta)
+
+        # We are moving fast, take the faster path
+        self._likely_homogeneous = (slowest < overhead_threshold)
+
+    def _slow_path_step_body(self, force=False):
+        # In the slow path, we don't make any assumption about how long
+        # iterations take. So on every iteration we must measure the time
+        self._measure_time()
+        if force or (self._display_timedelta >= self.time_thresh):
+            self.display_message()
+
     def step(self, inc=1, force=False):
         """
         Manually step progress update, either directly or by an increment.
@@ -507,160 +688,73 @@ class ProgIter(_TQDMCompat, _BackwardsCompat):
         """
         if not self.enabled:
             return
+
         self._iter_idx += inc
-        _between_idx = (self._iter_idx - self._now_idx)
-        if force or _between_idx >= self.freq:
-            self._update_measurements()
-            self._update_estimates()
-            self.display_message()
-
-    def _reset_internals(self):
-        """
-        Initialize all variables used in the internal state
-        """
-        # Prepare for iteration
-        if self.total is None:
-            self.total = _infer_length(self.iterable)
-        self._est_seconds_left = None
-        self._total_seconds = 0
-        self._between_time = 0
-        self._iter_idx = self.initial
-        self._last_idx = self.initial - 1
-        # now time is actually not right now
-        # now refers the the most recent measurement
-        # last refers to the measurement before that
-        self._now_idx = self.initial
-        self._now_time = 0
-        self._between_count = -1
-        self._max_between_time = -1.0
-        self._max_between_count = -1.0
-        self._iters_per_second = 0.0
-        self._update_message_template()
-
-    def start(self):  # nocover
-        """
-        Alias of :func:`ProgIter.begin`
-        """
-        return self.begin()
-
-    def begin(self):
-        """
-        Initializes information used to measure progress
-
-        This only needs to be used if this ProgIter is not wrapping an iterable.
-        Does nothing if the this ProgIter is disabled.
-
-        Returns:
-            ProgIter:
-                a chainable self-reference
-        """
-        if not self.enabled:
-            return
-
-        self._reset_internals()
-
-        self._tryflush()
-        self.display_message()
-
-        # Time progress was initialized
-        self._start_time = default_timer()
-        # Last time measures were updated
-        self._last_time  = self._start_time
-        self._now_idx = self._iter_idx
-        self._now_time = self._start_time
-
-        # use last few times to compute a more stable average rate
-        if self.eta_window is not None:
-            self._measured_times = collections.deque(
-                [], maxlen=self.eta_window)
-            self._measured_times.append((self._iter_idx, self._start_time))
-
-        # self._cursor_at_newline = True
-        self._cursor_at_newline = not self.clearline
-        self.started = True
-        self.finished = False
-        return self
-
-    def end(self):
-        """
-        Signals that iteration has ended and displays the final message.
-
-        This only needs to be used if this ProgIter is not wrapping an
-        iterable.  Does nothing if the this ProgIter object is disabled or has
-        already finished.
-        """
-        if not self.enabled or self.finished:
-            return
-        # Write the final progress line if it was not written in the loop
-        if self._iter_idx != self._now_idx:
-            self._update_measurements()
-            self._update_estimates()
-            self._est_seconds_left = 0
-            self.display_message()
-        self.ensure_newline()
-        self._cursor_at_newline = True
-        self.finished = True
+        self._slow_path_step_body(force=force)
 
     def _adjust_frequency(self):
         # Adjust frequency so the next print will not happen until
         # approximately `time_thresh` seconds have passed as estimated by
         # iter_idx.
-        eps = 1E-9
-        self._max_between_time = max(self._max_between_time,
-                                     self._between_time)
-        self._max_between_time = max(self._max_between_time, eps)
-        self._max_between_count = max(self._max_between_count,
-                                      self._between_count)
 
         # If progress was uniform and all time estimates were
         # perfect this would be the new freq to achieve self.time_thresh
-        new_freq = int(self.time_thresh * self._max_between_count /
-                       self._max_between_time)
+        eps = 1E-9
+        new_freq = int(self.time_thresh * self._measure_countdelta /
+                       max(eps, self._measure_timedelta))
         # But things are not perfect. So, don't make drastic changes
         rel_limit = self.rel_adjust_limit
         max_freq = int(self.freq * rel_limit)
         min_freq = int(self.freq // rel_limit)
         self.freq = max(min(new_freq, max_freq), min_freq, 1)
 
-    def _update_measurements(self):
+    def _measure_time(self):
         """
-        update current measurements and estimated of time and progress
+        Measures the current time and update info about how long we've been
+        waiting since the last iteration was displayed.
         """
-        self._last_idx = self._now_idx
-        self._last_time  = self._now_time
+        _prev_measurement = self._measurements[-1]
+        if _prev_measurement.idx == self._iter_idx:
+            # We already recorded this time measurement
+            # raise AssertionError("PROBABLY SHOULD NOT BE HERE")
+            return
 
-        self._now_idx = self._iter_idx
-        self._now_time = default_timer()
+        _curr_measurement = Measurement(self._iter_idx, self._timer())
 
-        self._between_time = self._now_time - self._last_time
-        self._between_count = self._now_idx - self._last_idx
-        self._total_seconds = self._now_time - self._start_time
+        self._curr_measurement = _curr_measurement
+        self._measurements.append(_curr_measurement)
 
-        # Record that measures were updated
+        self._measure_timedelta = _curr_measurement.time - _prev_measurement.time
+        self._measure_countdelta = _curr_measurement.idx - _prev_measurement.idx
+        self._total_seconds = _curr_measurement.time - self._start_time
+
+        self._display_timedelta = (self._curr_measurement.time -
+                                   self._display_measurement.time)
 
-    def _update_estimates(self):
         # Estimate rate of progress
         if self.eta_window is None:
-            self._iters_per_second = self._now_idx / self._total_seconds
+            self._iters_per_second = self._curr_measurement.idx / self._total_seconds
         else:
             # Smooth out rate with a window
-            self._measured_times.append((self._now_idx, self._now_time))
-            prev_idx, prev_time = self._measured_times[0]
-            self._iters_per_second =  ((self._now_idx - prev_idx) /
-                                       (self._now_time - prev_time))
+            oldest_idx, oldest_time = self._measurements[0]
+            latest_idx, latest_time = self._measurements[-1]
+            self._iters_per_second =  ((latest_idx - oldest_idx) /
+                                       (latest_time - oldest_time))
 
         if self.total is not None:
             # Estimate time remaining if total is given
-            iters_left = self.total - self._now_idx
+            iters_left = self.total - self._curr_measurement.idx
             est_eta = iters_left / self._iters_per_second
-            self._est_seconds_left  = est_eta
+            self._est_seconds_left = est_eta
 
-        # Adjust frequency if printing too quickly
-        # so progress does not slow down actual function
-        if self.adjust and (self._between_time < self.time_thresh or
-                            self._between_time > self.time_thresh * 2.0):
+        # Adjust frequency to stay within time_thresh
+        if self.adjust and (self._measure_timedelta < self.time_thresh or
+                            self._measure_timedelta > self.time_thresh * 2.0):
             self._adjust_frequency()
 
+        # Mark when our next measurement should be in "fast mode"
+        self._next_measure_idx = self._iter_idx + self.freq
+
     def _update_message_template(self):
         self._msg_fmtstr = self._build_message_template()
 
@@ -730,28 +824,36 @@ class ProgIter(_TQDMCompat, _BackwardsCompat):
         return parts
 
     def format_message(self):
+        """
+        Exists only for backwards compatibility.
+
+        See `format_message_parts` for more recent API.
+        """
+        return ''.join(self.format_message_parts())
+
+    def format_message_parts(self):
         r"""
         builds a formatted progres message with the current values.
         This contains the special characters needed to clear lines.
 
         Example:
             >>> self = ProgIter(clearline=False, show_times=False)
-            >>> print(repr(self.format_message()[1]))
+            >>> print(repr(self.format_message_parts()[1]))
             '    0/?... '
             >>> self.begin()
             >>> self.step()
-            >>> print(repr(self.format_message()[1]))
+            >>> print(repr(self.format_message_parts()[1]))
             ' 1/?... '
 
         Example:
             >>> self = ProgIter(chunksize=10, total=100, clearline=False,
             >>>                 show_times=False, microseconds=True)
             >>> # hack, microseconds=True for coverage, needs real test
-            >>> print(repr(self.format_message()[1]))
+            >>> print(repr(self.format_message_parts()[1]))
             ' 0.00% of 10x100... '
             >>> self.begin()
             >>> self.update()  # tqdm alternative to step
-            >>> print(repr(self.format_message()[1]))
+            >>> print(repr(self.format_message_parts()[1]))
             ' 1.00% of 10x100... '
         """
         from datetime import timedelta
@@ -774,7 +876,7 @@ class ProgIter(_TQDMCompat, _BackwardsCompat):
         if self.chunksize and self.total:
             msg = fmtstr.format(
                 desc=self.desc,
-                percent=self._now_idx / self.total * 100,
+                percent=self._curr_measurement.idx / self.total * 100,
                 rate=self._iters_per_second * self.chunksize,
                 rate_format='4.2f' if self._iters_per_second * self.chunksize > .001 else 'g',
                 eta=eta, total=total,
@@ -784,7 +886,7 @@ class ProgIter(_TQDMCompat, _BackwardsCompat):
         else:
             msg = fmtstr.format(
                 desc=self.desc,
-                iter_idx=self._now_idx,
+                iter_idx=self._curr_measurement.idx,
                 rate=self._iters_per_second,
                 rate_format='4.2f' if self._iters_per_second > .001 else 'g',
                 eta=eta, total=total,
@@ -802,7 +904,8 @@ class ProgIter(_TQDMCompat, _BackwardsCompat):
 
         Example:
             >>> # Unsafe version may write your message on the wrong line
-            >>> prog = ProgIter(range(3), show_times=False, freq=2, adjust=False)
+            >>> prog = ProgIter(range(3), show_times=False, freq=2, adjust=False,
+            ...                 time_thresh=0)
             >>> for n in prog:
             ...     print('unsafe message')
              0/3... unsafe message
@@ -812,7 +915,8 @@ class ProgIter(_TQDMCompat, _BackwardsCompat):
             >>> # apparently the safe version does this too.
             >>> print('---')
             ---
-            >>> prog = ProgIter(range(3), show_times=False, freq=2, adjust=False)
+            >>> prog = ProgIter(range(3), show_times=False, freq=2, adjust=False,
+            ...                 time_thresh=0)
             >>> for n in prog:
             ...     prog.ensure_newline()
             ...     print('safe message')
@@ -832,7 +936,11 @@ class ProgIter(_TQDMCompat, _BackwardsCompat):
         """
         Writes current progress to the output stream
         """
-        before, msg, after = self.format_message()
+        # When we make a display, ensure need to have a recent time measurement
+        if self._curr_measurement.idx != self._iter_idx:
+            self._measure_time()
+
+        before, msg, after = self.format_message_parts()
         msg_len = len(msg)  # TODO account for unicode
         if self.clearline:
             padding = self._prev_msg_len - msg_len
@@ -842,6 +950,9 @@ class ProgIter(_TQDMCompat, _BackwardsCompat):
         self._prev_msg_len = msg_len
         self._tryflush()
         self._cursor_at_newline = not self.clearline
+        self._display_measurement = self._curr_measurement
+        self._display_timedelta = 0
+        self._force_next_display = False
 
     def _tryflush(self):
         """ flush to the internal stream """
